#!/usr/bin/python

# gpsman-background, utility for GPSMan
# Copyright (C) 2005 Christian Mauduit <ufoot@ufoot.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by # the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# http://www.ufoot.org

#from numarray import *
#import numarray.linear_algebra as la

import numarray
import numarray.linear_algebra
import re
import sys
import Image
import math
import os

BG_SIZE=5000

class Transformer:
    """Class used to get a,b,c,d,e,f coefficients from 3 couples of points (xi,yi)=>(xi',yi'), the equation behind this being  (x',y')=(a*x+b*y+c,d*x+e*y+f)."""
    def __init__(self,points):
        """Creates the object initializing it with points data, a 3 dimensions array. First dim is for the source and target points. Second dim is for the different points. Third dim is to store x and y."""
        self.a=0.
        self.b=0.
        self.c=0.
        self.d=0.
        self.e=0.
        self.f=0.
        self.x1=float(points[0][0][0])
        self.y1=float(points[0][0][1])
        self.x2=float(points[0][1][0])
        self.y2=float(points[0][1][1])
        self.x3=float(points[0][2][0])
        self.y3=float(points[0][2][1])
        self.xp1=float(points[1][0][0])
        self.yp1=float(points[1][0][1])
        self.xp2=float(points[1][1][0])
        self.yp2=float(points[1][1][1])
        self.xp3=float(points[1][2][0])
        self.yp3=float(points[1][2][1])
        self.calc_abcdef()
    def calc_abcdef(self):
        """Calcultates the abcdef coefficients using matrix inversion."""
        matrix1=numarray.array([
            [self.x1,self.y1,1,0,0,0],
            [0,0,0,self.x1,self.y1,1],
            [self.x2,self.y2,1,0,0,0],
            [0,0,0,self.x2,self.y2,1],
            [self.x3,self.y3,1,0,0,0],
            [0,0,0,self.x3,self.y3,1]
            ])
        matrix2=numarray.linear_algebra.inverse(matrix1)
        #print matrix1
        #print matrix2
        vect=numarray.array([self.xp1,
                             self.yp1,
                             self.xp2,
                             self.yp2,
                             self.xp3,
                             self.yp3])
        abcdef=numarray.matrixmultiply(matrix2,vect)
        #print abcdef
        self.a=abcdef[0]
        self.b=abcdef[1]
        self.c=abcdef[2]
        self.d=abcdef[3]
        self.e=abcdef[4]
        self.f=abcdef[5]
    def calc(self,coord):
        """Gives the new coordinates of a point."""
        return (self.a*coord[0]+self.b*coord[1]+self.c,
                self.d*coord[0]+self.e*coord[1]+self.f)
    def get_abcdef(self):
        """Returns the coefficients."""
        return (self.a,
                self.b,
                self.c,
                self.d,
                self.e,
                self.f)

class CoordReader:
    """Reads coord files in .txt format associated with .jpg files."""
    def __init__(self,filename):
        """Opens and reads the file."""
        self.coords=[]
        repl=re.compile("(\.jpg|jpeg|gif|pnm|bmp|txt)$")
        f=open(repl.sub(".txt",filename),"r")
        format=re.compile("^\s*([\d|\.]+)\s*\,\s*([\d|\.]+)\s*\:\s*([\d|\.]+)\s*\,\s*([\d|\.]+)\s*$")
        for line in f.readlines():
            s=format.search(line)
            if (s):
                self.coords.append({"x": float(s.group(1)),
                                    "y": float(s.group(2)),
                                    "long": float(s.group(3)),
                                    "lat": -float(s.group(4))})  
        f.close()
        if (len(self.coords)!=3):
            raise "Bad coord file: %s." % filename
    def get_coords(self):
        return self.coords

class BackgroundBuilder:
    """Class used to build a georeferenced background from raw .jpg files associated with a plain .txt files."""
    def __init__(self,argv):
        """Construct object and parses arguments."""
        self.target=argv[1]
        self.sources=argv[2:]
        self.coords={}
    def calc_min_max_unit(self,source):
        """Calculates min and max GPS coordinates for a single image."""
        coords=CoordReader(source).get_coords()
        img=Image.open(source)
        tr=Transformer((((coords[0]['x'],
                          coords[0]['y'],),
                         (coords[1]['x'],
                          coords[1]['y'],),
                         (coords[2]['x'],
                          coords[2]['y'])),
                        ((coords[0]['long'],
                          coords[0]['lat']),
                         (coords[1]['long'],
                          coords[1]['lat']),
                         (coords[2]['long'],
                          coords[2]['lat']))))
        corners=(tr.calc((0,0)),
                 tr.calc((img.size[0],0)),
                 tr.calc((0,img.size[1])),
                 tr.calc((img.size[0],img.size[1])))
        xmin=min(corners[0][0],
                 corners[1][0],
                 corners[2][0],
                 corners[3][0])
        ymin=min(corners[0][1],
                 corners[1][1],
                 corners[2][1],
                 corners[3][1])
        xmax=max(corners[0][0],
                 corners[1][0],
                 corners[2][0],
                 corners[3][0])
        ymax=max(corners[0][1],
                 corners[1][1],
                 corners[2][1],
                 corners[3][1])
        return {'xmin':xmin, 'ymin':ymin, 'xmax':xmax, 'ymax':ymax}
    def calc_min_max_all(self):
        """Calculates min and max GPS coordinates for several images."""
        xmin=+math.pow(10,100);
        ymin=+math.pow(10,100);
        xmax=-math.pow(10,100);
        ymax=-math.pow(10,100);
        for source in self.sources:
            min_max=self.calc_min_max_unit(source)
            xmin=min(xmin,min_max['xmin'])
            ymin=min(ymin,min_max['ymin'])
            xmax=max(xmax,min_max['xmax'])
            ymax=max(ymax,min_max['ymax'])
        return {'xmin':xmin, 'ymin':ymin, 'xmax':xmax, 'ymax':ymax}
    def utm_to_degrees(self,(coord_utm)):
        cmd="invproj +proj=utm -f \"%.8f\""
        print "exec %(cmd)s" % {'cmd' : cmd}
        (fout,fin)=os.popen2(cmd)
        coord_utm="%(long)d %(lat)d\n" % {'long':int(coord_utm[0]),'lat':int(coord_utm[1])}
        fout.write(coord_utm)
        fout.close()
        coord_degrees=fin.read()
        r=re.compile("^\s*([\d|\.]+)\s+([\d|\.]+)\s*$")
        s=r.search(coord_degrees)
        return (float(s.group(1)),float(s.group(2)))
    def georef(self):
        """Georeferences the map by writing description text files."""
        coords=self.coords[self.sources[0]]
        p1=(coords[0]['long']*1000,-coords[0]['lat']*1000)
        p2=(coords[1]['long']*1000,-coords[1]['lat']*1000)
        p3=(coords[2]['long']*1000,-coords[2]['lat']*1000)
        p1=self.utm_to_degrees(p1)
        p2=self.utm_to_degrees(p2)
        p3=self.utm_to_degrees(p3)
        data={'long1':p1[0],
              'lat1':p1[1],
              'long2':p2[0],
              'lat2':p2[1],
              'long3':p3[0],
              'lat3':p3[1],
              'x1':coords[0]['target_x'],
              'y1':coords[0]['target_y'],
              'x2':coords[1]['target_x'],
              'y2':coords[1]['target_y'],
              'x3':coords[2]['target_x'],
              'y3':coords[2]['target_y'],
              'image':self.target}
        cmd="gpsman georef \"%(image)s\" Affine %(lat1).8f %(long1).8f %(lat2).8f %(long2).8f %(lat3).8f %(long3).8f \"WGS 84\" %(x1)d %(y1)d %(x2)d %(y2)d %(x3)d %(y3)d  UTM" % data
        print "exec %(cmd)s" % {'cmd' : cmd}
        (fout,fin)=os.popen2(cmd)
        conf=fin.read()
        repl=re.compile("(\.jpg|jpeg|gif|pnm|bmp)$")
        target_georef=repl.sub(".georef",self.target)
        f=open(target_georef,"w")
        f.write(conf)
        f.close()
    def build(self):
        """Builds the background."""
        print "Building %(target)s:" % {'target': self.target}
        self.min_max=self.calc_min_max_all()
        (w,h)=(self.min_max['xmax']-self.min_max['xmin'],
               self.min_max['ymax']-self.min_max['ymin'])
        self.coef=math.sqrt(BG_SIZE*BG_SIZE/(w*h))
        (w,h)=map(lambda x : x*self.coef, (w,h))
        target=Image.new("RGBA",(int(round(w)),int(round(h))),"#bfbfbf")
        for source in self.sources:
            print "Adding %(source)s..." % {'source': source}
            coords=CoordReader(source).get_coords()            
            min_max=self.calc_min_max_unit(source)
            xmin=(min_max['xmin']-self.min_max['xmin'])*self.coef
            ymin=(min_max['ymin']-self.min_max['ymin'])*self.coef
            xmax=(min_max['xmax']-self.min_max['xmin'])*self.coef
            ymax=(min_max['ymax']-self.min_max['ymin'])*self.coef
            tmp=Image.open(source)
            tr_data=((((coords[0]['long']-self.min_max['xmin'])*self.coef-xmin,
                       (coords[0]['lat']-self.min_max['ymin'])*self.coef-ymin),
                      ((coords[1]['long']-self.min_max['xmin'])*self.coef-xmin,
                       (coords[1]['lat']-self.min_max['ymin'])*self.coef-ymin),
                      ((coords[2]['long']-self.min_max['xmin'])*self.coef-xmin,
                       (coords[2]['lat']-self.min_max['ymin'])*self.coef-ymin)),
                     (((coords[0]['x'],
                        coords[0]['y'],),
                       (coords[1]['x'],
                        coords[1]['y'],),
                       (coords[2]['x'],
                        coords[2]['y']))))
            tr=Transformer(tr_data)
            for i in range(3):
                coords[i]['target_x']=int(round(tr_data[0][i][0]+xmin))
                coords[i]['target_y']=int(round(tr_data[0][i][1]+ymin))
            self.coords[source]=coords
            data=tr.get_abcdef()
            # We need to handle alpha channel in a separate way for
            # RGBA alpha channel is ignored when pasting
            alpha=Image.new("1",tmp.size,1)
            tmp=tmp.transform((int(round(xmax-xmin)),
                               int(round(ymax-ymin))),
                              Image.AFFINE,
                              data,
                              Image.BICUBIC)
            alpha=alpha.transform((int(round(xmax-xmin)),
                                   int(round(ymax-ymin))),
                                  Image.AFFINE,
                                  data,
                                  Image.NEAREST)
            target.paste(tmp,(int(round(xmin)),int(round(ymin))),alpha)
        target.save(self.target)
        self.georef()
        print "Done."
            
#t=Transformer((((0,0),(0,1.1),(1,0)),((12,23),(-34,22),(43,56))))
#print t.calc((0,0))
#print t.calc((0,1))
#print t.calc((1,1,0))

bb=BackgroundBuilder(sys.argv)
bb.build()

